// Copyright © 2023 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#include <iostream>
#include <filesystem>
#include <thread>
#include <atomic>
#include <map>
#include <getopt.h>
#include "es_npu_interface.h"
#include "es_sys_memory.h"
#include "sample_npu_utils.h"
#include "EsVdec.h"
#include "EsHwPreProcess.h"
#include "EsPostProcess.h"
#include "sample_npu_comm.h"

static const std::string DEFAULT_INPUT_DIR(
    "/opt/eswin/sample-code/npu_sample/npu_runtime_sample/models/yolov3/input/pictures/");
static const std::string DEFAULT_MODEL_DIR("/opt/eswin/sample-code/npu_sample/npu_runtime_sample/models/yolov3/");
static const ES_S32 DEFAULT_NUMBERS = 4;

enum {
    SAMPLE_SYNC = 1,
    SAMPLE_ASYNC,
    SAMPLE_MULTI_CONTEXT,
    SAMPLE_MULTI_STREAM,
    SAMPLE_MULTI_MODEL,
    SAMPLE_COMPOSITE_MODEL,
    SAMPLE_D2D,
    SAMPLE_PEPELINE,
};

static struct option long_option[] = {
    {"help", no_argument, NULL, 'h'},
    {"sample", required_argument, NULL, 's'},
    {"model", required_argument, NULL, 'm'},
    {"input", required_argument, NULL, 'i'},
    {"output", required_argument, NULL, 'o'},
    {"pre_process", required_argument, NULL, 'p'},
    {"post_process", required_argument, NULL, 'q'},
    {"classes", required_argument, NULL, 't'},
    {"numbers", required_argument, NULL, 't'},
    {"device", required_argument, NULL, 'd'},
};

ES_S32 SAMPLE_NPU_Sync(ES_U16 devId, ES_U32 modelId, std::vector<std::vector<std::string>> &inputFiles,
                       const std::string &outputDir, std::shared_ptr<IVdec> esVdec,
                       std::shared_ptr<IPreprocess> preProcess, std::shared_ptr<EsPostProcess> postProcess)
{
    return SAMPLE_NPU_COMM_SubmitSync(devId, modelId, inputFiles, outputDir, esVdec, preProcess, postProcess);
}

ES_S32 SAMPLE_NPU_ASync(ES_U16 devId, ES_U32 modelId, std::vector<std::vector<std::string>> &inputFiles,
                        std::string &outputDir, std::shared_ptr<IVdec> esVdec, std::shared_ptr<IPreprocess> preProcess,
                        std::shared_ptr<EsPostProcess> postProcess)
{
    ES_S32 ret = ES_SUCCESS, err = ES_SUCCESS;
    ES_U32 taskId;
    std::atomic<ES_S32> completeNum(0);
    SampleStateInfo sampleInfo;
    npu_stream stream;
    npu_context context;

    ret = ES_NPU_CreateContext(&context, devId);
    if (ret != ES_SUCCESS) {
        printf("ES_NPU_CreateContext failed, ret: 0x%x\n", ret);
        return ES_FAILURE;
    }

    ret = ES_NPU_CreateStream(&stream);
    if (ret != ES_SUCCESS) {
        printf("ES_NPU_CreateStream failed, ret: 0x%x\n", ret);
        goto destroy_context;
    }

    ret = SAMPLE_NPU_COMM_SubmitAsync(devId, modelId, stream, inputFiles, outputDir, esVdec, preProcess, postProcess);
    if (ret != ES_SUCCESS) {
        printf("SAMPLE_NPU_COMM_SubmitAsync failed\n");
    }

    err = ES_NPU_DestroyStream(stream);
    if (err != ES_SUCCESS) {
        printf("ES_NPU_DestroyStream failed, ret: 0x%x\n", ret);
    }

destroy_context:
    err = ES_NPU_DestroyContext(context);
    if (err != ES_SUCCESS) {
        printf("ES_NPU_DestroyContext failed, ret: 0x%x\n", ret);
    }

    return ret | err;
}

ES_S32 SAMPLE_NPU_MultiContext(ES_U16 devId, ES_S32 nums, ES_U32 modelId,
                               std::vector<std::vector<std::string>> &inputFiles, std::string &outputDir,
                               std::shared_ptr<IVdec> esVdec, std::shared_ptr<IPreprocess> preProcess,
                               std::shared_ptr<EsPostProcess> postProcess)
{
    ES_S32 ret;
    std::vector<npu_context> contextVec;
    ES_S32 contextId = 0;

    for (contextId = 0; contextId < nums; contextId++) {
        npu_context context;
        ret = ES_NPU_CreateContext(&context, devId);
        if (ret != ES_SUCCESS) {
            printf("ES_NPU_CreateContext, ret: 0x%x\n", ret);
            goto destroy_context;
        }
        contextVec.push_back(context);
    }

    for (auto &context : contextVec) {
        npu_stream stream;
        ret = ES_NPU_SetCurrentContext(context);
        if (ret != ES_SUCCESS) {
            printf("ES_NPU_CreateContext, ret: 0x%x\n", ret);
            goto destroy_context;
        }

        ret = ES_NPU_CreateStream(&stream);
        if (ret != ES_SUCCESS) {
            printf("ES_NPU_CreateStream, ret: 0x%x\n", ret);
            goto destroy_context;
        }

        ret =
            SAMPLE_NPU_COMM_SubmitAsync(devId, modelId, stream, inputFiles, outputDir, esVdec, preProcess, postProcess);
        if (ret != ES_SUCCESS) {
            printf("SAMPLE_NPU_COMM_SubmitAsync failed\n");
        }

        ret |= ES_NPU_DestroyStream(stream);
        if (ret != ES_SUCCESS) {
            goto destroy_context;
        }
    }

destroy_context:
    for (auto &context : contextVec) {
        ret |= ES_NPU_DestroyContext(context);
    }

    return ret;
}

ES_S32 SAMPLE_NPU_MultiStream(ES_U16 devId, ES_S32 nums, ES_U32 modelId,
                              std::vector<std::vector<std::string>> &inputFiles, std::string &outputDir,
                              std::shared_ptr<IVdec> esVdec, std::shared_ptr<IPreprocess> preProcess,
                              std::shared_ptr<EsPostProcess> postProcess)
{
    ES_S32 ret;
    npu_context context;
    ES_S32 streamId = 0;
    std::vector<npu_stream> streamVec;

    ret = ES_NPU_CreateContext(&context, devId);
    if (ret != ES_SUCCESS) {
        printf("ES_NPU_CreateContext, ret: 0x%x\n", ret);
        return ret;
    }

    for (streamId = 0; streamId < nums; streamId++) {
        npu_stream stream;
        ret = ES_NPU_CreateStream(&stream);
        if (ret != ES_SUCCESS) {
            printf("ES_NPU_CreateStream, ret: 0x%x\n", ret);
            goto destroy_stream;
        }
        streamVec.push_back(stream);
    }

    for (auto &stream : streamVec) {
        ret =
            SAMPLE_NPU_COMM_SubmitAsync(devId, modelId, stream, inputFiles, outputDir, esVdec, preProcess, postProcess);
        if (ret != ES_SUCCESS) {
            printf("SAMPLE_NPU_COMM_SubmitAsync failed\n");
            goto destroy_stream;
        }
    }

destroy_stream:
    for (auto &stream : streamVec) {
        ret |= ES_NPU_DestroyStream(stream);
    }

    return ret;
}

ES_S32 SAMPLE_NPU_MultiModel(ES_U16 devId, std::vector<std::string> &modelPaths, std::vector<std::string> &inputsDir,
                             std::vector<std::string> &outputDirs,
                             std::map<ES_U16, std::map<ES_U16, std::shared_ptr<IVdec>>> &esVdec,
                             std::map<ES_U16, std::map<ES_U16, std::shared_ptr<IPreprocess>>> &preProcess,
                             std::map<ES_U16, std::map<ES_U16, std::shared_ptr<EsPostProcess>>> &postProcess)
{
    ES_S32 ret;
    ES_U32 modelId;
    std::vector<ES_U32> modelIdVec;
    std::vector<std::string> inputs;
    std::vector<std::vector<std::string>> inputFiles;

    for (auto &modelPath : modelPaths) {
        ret = ES_NPU_LoadModelFromFile(&modelId, modelPath.c_str());
        if (ret != ES_SUCCESS) {
            printf("ES_NPU_LoadModelFromFile failed, ret: 0x%x\n", ret);
            goto unload_model;
        }
        modelIdVec.push_back(modelId);
    }

    for (ES_S32 i = 0; i < modelIdVec.size(); i++) {
        std::string outputDir;
        inputFiles.clear();
        inputs = listSubDirectory(inputsDir[i]);
        for (auto &input : inputs) {
            inputFiles.push_back(listSubDirectory(input));
        }
        if (inputFiles.empty() || inputFiles.front().empty()) {
            printf("No input file found, inputDir: %s\n", inputsDir[i].c_str());
            ret = ES_FAILURE;
            goto unload_model;
        }

        if (outputDirs.size() > i) {
            outputDir = outputDirs[i];
        }

        ret = SAMPLE_NPU_COMM_SubmitSync(devId, modelIdVec[i], inputFiles, outputDir, esVdec[i][devId],
                                         preProcess[i][devId], postProcess[i][devId]);
        if (ret != ES_SUCCESS) {
            printf("SAMPLE_NPU_COMM_SubmitSync failed\n");
            goto unload_model;
        }
    }

unload_model:
    for (auto &id : modelIdVec) {
        ret |= ES_NPU_UnloadModel(id);
    }

    return ret;
}

ES_S32 SAMPLE_NPU_CompositeModel(ES_U32 modelId, ES_U16 devId, std::vector<std::vector<std::string>> &inputFiles,
                                 std::string &outputDir, std::shared_ptr<IVdec> esVdec,
                                 std::shared_ptr<IPreprocess> preProcess, std::shared_ptr<EsPostProcess> postProcess)
{
    ES_S32 ret, err;
    npu_stream stream;
    npu_context context;
    NPU_FLEXIBLE_TASK_ATTR_S attr;

    ret = ES_NPU_CreateContext(&context, devId);
    if (ret != ES_SUCCESS) {
        printf("(%s, %d)ES_NPU_CreateContext failed, ret: 0x%x\n", __func__, __LINE__, ret);
        return ES_FAILURE;
    }

    attr.timeOut = 100;
    ret = ES_NPU_SetFlexibleTaskAttr(modelId, &attr);
    if (ret != ES_SUCCESS) {
        printf("(%s, %d)ES_NPU_SetFlexibleTaskAttr failed, ret: 0x%x\n", __func__, __LINE__, ret);
        goto destroy_context;
    }

    ret = ES_NPU_CreateStream(&stream);
    if (ret != ES_SUCCESS) {
        printf("(%s, %d)ES_NPU_CreateStream failed, ret: 0x%x\n", __func__, __LINE__, ret);
        goto destroy_context;
    }

    ret = SAMPLE_NPU_COMMON_SubmitUserTask(modelId, stream, inputFiles, outputDir, esVdec, preProcess, postProcess);
    if (ret != ES_SUCCESS) {
        printf("(%s, %d)SAMPLE_NPU_COMMON_SubmitUserTask failed,  ret: 0x%x\n", __func__, __LINE__, ret);
    }

    err = ES_NPU_DestroyStream(stream);
    if (err != ES_SUCCESS) {
        printf("ES_NPU_DestroyStream failed, ret: 0x%x\n", ret);
    }

destroy_context:
    err = ES_NPU_DestroyContext(context);
    if (err != ES_SUCCESS) {
        printf("ES_NPU_DestroyContext failed, ret: 0x%x\n", ret);
    }

    return err | ret;
}

ES_S32 SAMPLE_NPU_Pipeline(ES_U32 modelId, ES_U16 devId, std::vector<std::vector<std::string>> &inputFiles,
                           std::string &outputDir, std::shared_ptr<IVdec> esVdec,
                           std::shared_ptr<IPreprocess> preProcess, std::shared_ptr<EsPostProcess> postProcess)
{
    ES_S32 ret, err;
    npu_stream stream;
    npu_context context;
    NPU_FLEXIBLE_TASK_ATTR_S attr;

    ret = ES_NPU_CreateContext(&context, devId);
    if (ret != ES_SUCCESS) {
        printf("(%s, %d)ES_NPU_CreateContext failed, ret: 0x%x\n", __func__, __LINE__, ret);
        return ES_FAILURE;
    }

    attr.timeOut = 100;
    ret = ES_NPU_SetFlexibleTaskAttr(modelId, &attr);
    if (ret != ES_SUCCESS) {
        printf("(%s, %d)ES_NPU_SetFlexibleTaskAttr failed, ret: 0x%x\n", __func__, __LINE__, ret);
        goto destroy_context;
    }

    ret = ES_NPU_CreateStream(&stream);
    if (ret != ES_SUCCESS) {
        printf("(%s, %d)ES_NPU_CreateStream failed, ret: 0x%x\n", __func__, __LINE__, ret);
        goto destroy_context;
    }

    ret = SAMPLE_NPU_COMMON_SubmitPipelineTask(modelId, stream, inputFiles, outputDir, esVdec, preProcess, postProcess);
    if (ret != ES_SUCCESS) {
        printf("(%s, %d)SAMPLE_NPU_COMMON_SubmitUserTask failed,  ret: 0x%x\n", __func__, __LINE__, ret);
    }

    err = ES_NPU_DestroyStream(stream);
    if (err != ES_SUCCESS) {
        printf("ES_NPU_DestroyStream failed, ret: 0x%x\n", ret);
    }

destroy_context:
    err = ES_NPU_DestroyContext(context);
    if (err != ES_SUCCESS) {
        printf("ES_NPU_DestroyContext failed, ret: 0x%x\n", ret);
    }

    return err | ret;
}

ES_S32 SAMPLE_NPU_D2D(std::vector<std::vector<std::string>> &inputFiles, std::string &outputDir, std::string &modelPath,
                      std::map<ES_U16, std::shared_ptr<IVdec>> &esVdec,
                      std::map<ES_U16, std::shared_ptr<IPreprocess>> &preProcess,
                      std::map<ES_U16, std::shared_ptr<EsPostProcess>> &postProcess)
{
    ES_U16 deviceNum = 0;
    std::vector<std::thread> threadVec;
    ES_S32 ret;
    std::mutex mtx;

    ret = ES_NPU_GetNumDevices(&deviceNum);
    for (ES_U16 die = 0; die < deviceNum; die++) {
        threadVec.emplace_back(std::thread(SAMPLE_NPU_COMM_D2DProcessTrd, die, modelPath, std::ref(inputFiles),
                                           outputDir, esVdec[die], preProcess[die], postProcess[die], std::ref(mtx)));
    }

    for (auto &trd : threadVec) {
        if (trd.joinable()) {
            trd.join();
        }
    }

    return ES_SUCCESS;
}

std::vector<std::vector<std::string>> parseInputFiles(std::string &inputDir)
{
    std::vector<std::vector<std::string>> inputFiles;

    std::vector<std::string> inputVec = listSubDirectory(inputDir);
    for (auto &input : inputVec) {
        inputFiles.push_back(listSubDirectory(input));
    }

    return inputFiles;
}

ES_VOID printUsage()
{
    printf("Usage:\n");
    printf("-h, --help                display help info\n");
    printf("-s, --sample=type         set sample type(\n"
           "                          1.SAMPLE_SYNC\n"
           "                          2.SAMPLE_ASYNC\n"
           "                          3.SAMPLE_MULTI_CONTEXT\n"
           "                          4.SAMPLE_MULTI_STREAM\n"
           "                          5.SAMPLE_MULTI_MODEL\n"
           "                          6.SAMPLE_COMPOSITE_MODEL\n"
           "                          7.SAMPLE_D2D\n"
           "                          8.SAMPLE_PEPELINE)\n");
    printf("-m, --model=dir           set model dirs.(Multiple model dirs are separated by commas)\n");
    printf("-i, --input=dir           set input dirs.(Multiple input dirs are separated by commas)\n");
    printf("-o, --output=dir          set output dirs.(Multiple input dirs are separated by commas)\n");
    printf("-p, --pre_process=path    set pre-process config file.(Multiple config file are separated by commas)\n");
    printf("-q, --post_process=path   set post-process config file.(Multiple config file are separated by commas)\n");
    printf("-t, --classify=path       set classify file.(Multiple classify file are separated by commas)\n");
    printf("-n, --numbers=n           set numbers of context/stream.\n");
}

int main(int argc, char **argv)
{
    ES_U16 deviceId = 0;
    ES_U16 deviceNum = 0;
    ES_S32 numbers = DEFAULT_NUMBERS;
    ES_S32 sampleType = 0;
    ES_S32 ret;
    std::vector<std::string> preConfigFileVec;
    std::vector<std::string> postConfigFileVec;
    std::vector<std::string> classifyFileVec;
    std::vector<std::string> modelDirVec;
    std::vector<std::string> inputDirVec;
    std::vector<std::string> outputDirVec;
    std::map<ES_U16, std::map<ES_U16, std::shared_ptr<IVdec>>> esVdecMap;
    std::map<ES_U16, std::map<ES_U16, std::shared_ptr<IPreprocess>>> preProcessMap;
    std::map<ES_U16, std::map<ES_U16, std::shared_ptr<EsPostProcess>>> postProcessMap;
    std::vector<std::vector<std::string>> inputs;
    std::vector<std::string> modelPaths;
    ES_S32 c = 0;
    ES_S32 option_index = 0;

    while (1) {
        c = getopt_long(argc, argv, "hs:m:i:o:p:q:t:n:d:", long_option, &option_index);
        if (c == -1) break;
        switch (c) {
            case 'h':
                printUsage();
                return 0;
            case 's':
                sampleType = atoi(optarg);
                break;
            case 'm':
                modelDirVec = split(std::string(optarg), ',');
                break;
            case 'i':
                inputDirVec = split(std::string(optarg), ',');
                break;
            case 'o':
                outputDirVec = split(std::string(optarg), ',');
                break;
            case 'p':
                preConfigFileVec = split(std::string(optarg), ',');
                break;
            case 'q':
                postConfigFileVec = split(std::string(optarg), ',');
                break;
            case 't':
                classifyFileVec = split(std::string(optarg), ',');
                break;
            case 'n':
                numbers = atoi(optarg);
                break;
            case 'd':
                deviceId = atoi(optarg);
                break;
            default:
                printf("arg parse error, option is invalid: %c\n", c);
                printUsage();
                return 0;
        }
    }

    if (optind < argc) {
        printf("arg parse error, optind: %d\n", optind);
        printUsage();
        return 0;
    }

    if (inputDirVec.empty()) {
        inputDirVec.push_back(DEFAULT_INPUT_DIR);
    }

    if (modelDirVec.empty()) {
        modelDirVec.push_back(DEFAULT_MODEL_DIR);
    }

    if (preConfigFileVec.empty()) {
        for (auto &entry : modelDirVec) {
            std::string preConfig;
            preConfig = extractConfigFiles(entry, "pre_process.json");
            if (preConfig.empty()) {
                printf("Can not find pre_process config file in %s\n", entry.c_str());
                return 0;
            }
            preConfigFileVec.push_back(preConfig);
        }
    }

    if (postConfigFileVec.empty()) {
        for (auto &entry : modelDirVec) {
            std::string postConfig;
            postConfig = extractConfigFiles(entry, "post_process.json");
            if (postConfig.empty()) {
                printf("Can not find pre_process config file in %s\n", entry.c_str());
                return 0;
            }
            postConfigFileVec.push_back(postConfig);
        }
    }

    if (classifyFileVec.empty()) {
        for (auto &entry : modelDirVec) {
            std::string classesConfig;
            classesConfig = extractConfigFiles(entry, "classes.txt");
            classifyFileVec.push_back(classesConfig);
        }
    }

    if (modelDirVec.size() != inputDirVec.size() || modelDirVec.size() != preConfigFileVec.size() ||
        modelDirVec.size() != postConfigFileVec.size() || modelDirVec.size() != classifyFileVec.size()) {
        printf("Numbers of models is not equal to numbers of config or numbers of input\n");
        return 0;
    }

    ret = ES_NPU_GetNumDevices(&deviceNum);
    if (ret != ES_SUCCESS) {
        printf("ES_NPU_GetNumDevices failed\n");
        return 0;
    }

    for (ES_S32 modelIdx = 0; modelIdx < modelDirVec.size(); modelIdx++) {
        // esVdecVec[modelIdx] = {};
        // preProcessVec[modelIdx] = {};
        // postProcessVec[modelIdx] = {};
        for (ES_U16 dev = 0; dev < deviceNum; dev++) {
            esVdecMap[modelIdx].insert({dev, std::make_shared<EsVdec>(dev)});
            preProcessMap[modelIdx].insert({dev, std::make_shared<EsHwPreProcess>(dev)});
            postProcessMap[modelIdx].insert({dev, std::make_shared<EsPostProcess>(dev)});
        }
    }

    for (ES_S32 modelIdx = 0; modelIdx < modelDirVec.size(); modelIdx++) {
        for (ES_U16 dev = 0; dev < deviceNum; dev++) {
            if (esVdecMap[modelIdx][dev]->initVdec(deviceNum) != ES_SUCCESS) {
                printf("Error: Init video decoder failed.\n");
                return ES_FAILURE;
            }

            if (!preProcessMap[modelIdx][dev]->initializeFromJson(deviceId, preConfigFileVec[modelIdx])) {
                printf("PreProcess initializeFromJson failed, idx: %d\n", modelIdx);
                return 0;
            }

            if (!postProcessMap[modelIdx][dev]->parseArgs(postConfigFileVec[modelIdx])) {
                printf("PostProcess parseArgs failed, idx: %d\n", modelIdx);
                return 0;
            }

            PREPROCERSS_PARAMS &preProcessParams = preProcessMap[modelIdx][dev]->getParams();
            // Set the post-processing keep ratio from pre-processing
            POSTPROCESS_PARAMS postProcessParams;
            postProcessParams.keepRatio = preProcessParams.keepRatio;
            postProcessParams.keepLongSide = preProcessParams.keepLongSide;
            postProcessParams.paddingPosition = preProcessParams.paddingPosition;
            postProcessMap[modelIdx][dev]->setParams(postProcessParams);

            if (postProcessMap[modelIdx][dev]->getProcessType() == EsPostProcess::CLASSIFY ||
                postProcessMap[modelIdx][dev]->getProcessType() == EsPostProcess::DETECTION_OUT) {
                if (!postProcessMap[modelIdx][dev]->parseClassification(classifyFileVec[modelIdx])) {
                    printf("PostProcess parse classification text failed, idx: %d\n", modelIdx);
                    return 0;
                }
            }
        }
    }

    ret = extractModelFiles(modelDirVec, modelPaths);
    if (ret != ES_SUCCESS) {
        printf("extractModelFiles failed\n");
        return 0;
    }

    if (sampleType == SAMPLE_D2D) {
        for (ES_S32 modelIdx = 0; modelIdx < modelDirVec.size(); modelIdx++) {
            std::string outputDir;

            if (outputDirVec.size() > modelIdx) {
                outputDir = outputDirVec[modelIdx];
            }
            inputs = parseInputFiles(inputDirVec[modelIdx]);
            ret = SAMPLE_NPU_D2D(inputs, outputDir, modelDirVec[modelIdx], esVdecMap[modelIdx], preProcessMap[modelIdx],
                                 postProcessMap[modelIdx]);
        }
    } else {
        if (deviceId >= deviceNum) {
            printf("Invalid deviceId(%d) >= deviceNum(%d)\n", deviceId, deviceNum);
            return 0;
        }

        ret = SAMPLE_NPU_COMM_InitDevice(deviceId);
        if (ret != ES_SUCCESS) {
            printf("SAMPLE_NPU_COMM_InitDevice failed\n");
            return 0;
        }

        if (sampleType != SAMPLE_MULTI_MODEL) {
            for (ES_S32 modelIdx = 0; modelIdx < modelDirVec.size(); modelIdx++) {
                // Parse input picture & Load model
                ES_U32 modelId;
                std::string outputDir;
                inputs.clear();
                std::vector<std::string> inputVec = listSubDirectory(inputDirVec[modelIdx]);
                for (auto &input : inputVec) {
                    inputs.push_back(listSubDirectory(input));
                }
                if (inputs.empty() || inputs.front().empty()) {
                    goto releaseDevice;
                }
                if (outputDirVec.size() > modelIdx) {
                    outputDir = outputDirVec[modelIdx];
                }

                if (sampleType != SAMPLE_COMPOSITE_MODEL) {
                    ret = ES_NPU_LoadModelFromFile(&modelId, modelPaths[modelIdx].c_str());
                    if (ret != ES_SUCCESS) {
                        printf("Load model failed, ret: 0x%x\n", ret);
                        goto releaseDevice;
                    }
                    switch (sampleType) {
                        case SAMPLE_SYNC: {
                            ret =
                                SAMPLE_NPU_Sync(deviceId, modelId, inputs, outputDir, esVdecMap[modelIdx][deviceId],
                                                preProcessMap[modelIdx][deviceId], postProcessMap[modelIdx][deviceId]);
                            break;
                        }
                        case SAMPLE_ASYNC: {
                            ret =
                                SAMPLE_NPU_ASync(deviceId, modelId, inputs, outputDir, esVdecMap[modelIdx][deviceId],
                                                 preProcessMap[modelIdx][deviceId], postProcessMap[modelIdx][deviceId]);
                            break;
                        }
                        case SAMPLE_MULTI_CONTEXT: {
                            ret = SAMPLE_NPU_MultiContext(
                                deviceId, numbers, modelId, inputs, outputDir, esVdecMap[modelIdx][deviceId],
                                preProcessMap[modelIdx][deviceId], postProcessMap[modelIdx][deviceId]);
                            break;
                        }
                        case SAMPLE_MULTI_STREAM: {
                            ret = SAMPLE_NPU_MultiStream(
                                deviceId, numbers, modelId, inputs, outputDir, esVdecMap[modelIdx][deviceId],
                                preProcessMap[modelIdx][deviceId], postProcessMap[modelIdx][deviceId]);
                            break;
                        }
                        case SAMPLE_PEPELINE: {
                            ret = ES_NPU_LoadCompositeModel(&modelId, modelDirVec[modelIdx].c_str());
                            if (ret != ES_SUCCESS) {
                                printf("Load composite model failed, ret: 0x%x\n", ret);
                                goto releaseDevice;
                            }

                            ret = SAMPLE_NPU_Pipeline(modelId, deviceId, inputs, outputDir,
                                                      esVdecMap[modelIdx][deviceId], preProcessMap[modelIdx][deviceId],
                                                      postProcessMap[modelIdx][deviceId]);
                            ES_NPU_UnloadCompositeModel(modelId);
                            break;
                        }
                        default:
                            printf("Invalid sampleType: %d\n", sampleType);
                            printUsage();
                            return 0;
                    }
                    if (sampleType != SAMPLE_PEPELINE) ES_NPU_UnloadModel(modelId);
                } else {
                    ret = ES_NPU_LoadCompositeModel(&modelId, modelDirVec[modelIdx].c_str());
                    if (ret != ES_SUCCESS) {
                        printf("Load composite model failed, ret: 0x%x\n", ret);
                        goto releaseDevice;
                    }
                    ret = SAMPLE_NPU_CompositeModel(modelId, deviceId, inputs, outputDir, esVdecMap[modelIdx][deviceId],
                                                    preProcessMap[modelIdx][deviceId],
                                                    postProcessMap[modelIdx][deviceId]);
                    ES_NPU_UnloadCompositeModel(modelId);
                }
                if (ret != ES_SUCCESS) {
                    printf("sample_npu: %d failed, modelPath: %s\n", sampleType, modelDirVec[modelIdx].c_str());
                    goto releaseDevice;
                }
            }
        } else {
            ret = SAMPLE_NPU_MultiModel(deviceId, modelPaths, inputDirVec, outputDirVec, esVdecMap, preProcessMap,
                                        postProcessMap);
            if (ret != ES_SUCCESS) {
                printf("sample_npu: %d failed\n", sampleType);
            }
        }
releaseDevice:
        SAMPLE_NPU_COMM_ReleaseDevice(deviceId);
    }

    return 0;
}