// Copyright © 2024 ESWIN. All rights reserved.
//
// Beijing ESWIN Computing Technology Co., Ltd and its affiliated companies ("ESWIN") retain
// all intellectual property and proprietary rights in and to this software. Except as expressly
// authorized by ESWIN, no part of the software may be released, copied, distributed, reproduced,
// modified, adapted, translated, or created derivative work of, in whole or in part.

#ifndef _IVDEC_H
#define _IVDEC_H

#include <string>
#include <queue>
#include <thread>
#include <mutex>
#include <atomic>
#include <condition_variable>
#include <stdexcept>
#include "sample_comm.h"

typedef struct {
    uint8_t *data;
    uint32_t len;
    std::string name;
} JPEG_DATA;

typedef struct {
    VIDEO_FRAME_INFO_S videoInfoFrame;
    std::string name;
} VDEC_VIDEO_FRAME;

class IVdec
{
 public:
    IVdec() {}
    virtual ~IVdec() {}

 public:
    /**
     * @brief Initialize the video decoder with the specified device ID.
     *
     * @return Returns 0 if initialization is successful, a negative error code otherwise.
     */
    virtual int32_t initVdec(int32_t dienum) = 0;
    /**
     * @brief Decode a JPEG image into a video frame.
     *
     * @param jpg: The JPEG data to be decoded.
     * @return Returns 0 if decoding is successful, a negative error code otherwise.
     */
    virtual int32_t decodeJpeg(const JPEG_DATA &jpg) = 0;
    /**
     * @brief Retrieve the next video frame from the decoder.
     *
     * @param frame: The decoded video frame to be filled.
     * @return Returns true if a video frame is successfully retrieved, false otherwise.
     */
    virtual bool getVideoFrame(VDEC_VIDEO_FRAME &frame) = 0;
    /**
     * @brief Release the memory or resources used by the video frame.
     *
     * @param frame: The video frame to be released.
     * @return None.
     */
    virtual void releaseFrame(VDEC_VIDEO_FRAME &frame) = 0;
    /**
     * @brief Flush the video decoder, clearing any buffered data.
     *
     * @return None.
     */
    virtual void flush() = 0;
    /**
     * @brief Deinitialize the video decoder and release associated resources.
     *
     * @return None.
     */
    virtual void deInitVdec() = 0;
};

#endif  // _IVDEC_H